from pathlib import Path
from typing import List
import pandas as pd
import numpy as np
from scipy.stats import logistic as scipy_logistic  # type: ignore
import psychoanalyze as pa
from itertools import accumulate


def subjects(n_subjects):
    return list("ABCDEFG"[:n_subjects])


def construct_index(subjects: List[str], days: List[int], x: List[float]):
    levels = [days, x]
    names = ["Day", "x"]
    if subjects:
        levels = [subjects] + levels  # type: ignore
        names = ["Subject"] + names
    return pd.MultiIndex.from_product(levels, names=names)


def generate_outcomes(n_trials_per_stim_level, index, threshold, scale):
    return np.random.binomial(
        n_trials_per_stim_level,
        scipy_logistic.cdf(index.get_level_values("x"), threshold, scale),
        len(index),
    )


def psych(hits, n_trials_per_stim_level, index, y):
    df = pd.DataFrame(
        {
            "Hits": hits,
            y: hits / n_trials_per_stim_level,
            "n": [n_trials_per_stim_level] * len(index),
        },
        index=index,
    )
    df["Hit Rate"] = pa.blocks.hit_rate
    return df


def logistic(threshold=0, scale=1, gamma=0, lambda_=0):
    x = np.linspace(scipy_logistic.ppf(0.01), scipy_logistic.ppf(0.99), 100)
    index = pd.Index(x, name="x")
    return pd.Series(
        gamma + (1 - gamma - lambda_) * scipy_logistic.cdf(x, threshold, scale),
        index=index,
        name="Hit Rate",
    )


def mu(fit):
    df = fit.loc["mu", ["5%", "50%", "95%"]]  # type: ignore
    return df.T


def transform_errors(df):
    df["err+"] = df["95%"] - df["50%"]
    df["err-"] = df["50%"] - df["5%"]
    return df.drop(columns=["95%", "5%"])


def reshape_fit_results(fit: pd.DataFrame, x: pd.Index, y: str) -> pd.DataFrame:
    rows = [f"{y}[{i}]" for i in range(1, len(x) + 1)]
    df = fit.loc[
        rows,  # row eg 'p[1]:p[8]'
        ["5%", "50%", "95%"],  # col
    ]
    df = transform_errors(df)
    df = df.rename(columns={"50%": y})
    df.index = x
    return df


def generate_animation_curves():
    n_blocks = 10
    n_trials_per_level_per_block = 10
    df = pd.concat(
        list(
            accumulate(
                [
                    pa.blocks.generate(n_trials_per_level_per_block)
                    for _ in range(n_blocks)
                ]
            )
        )
    )
    df["Hit Rate"] = pa.blocks.hit_rate
    return df


def filter(df, dim):
    return df[df["Dimension"] == dim]


def load(data_dir=Path("data"), monkey=None, day=None):
    return {
        "Sessions": pa.sessions.load_cached(data_dir),
        "Subjects": pa.subjects.load(data_dir),
        "Blocks": pa.blocks.load(data_dir, monkey=monkey, day=day),
        "Points": pa.points.load(data_dir),
    }



import plotly.express as px  # type: ignore
import pandas as pd
import psychoanalyze as pa
from plotly import graph_objects as go

axis_settings = {
    "ticks": "outside",
    # "rangemode": "tozero",
    "showgrid": False,
    "showline": True,
    "zeroline": False,
    "title": {"font": {"size": 12, "family": "Arial"}},
}

template = go.layout.Template(
    layout=go.Layout(
        template="plotly_white",
        xaxis=axis_settings,
        yaxis=axis_settings,
        colorway=["#e41a1c", "#377eb8", "#4daf4a"],
        title={"font": {"size": 16, "family": "Arial"}},
        legend={"yanchor": "top", "y": 1, "xanchor": "left", "x": 0.98},
    ),
)

colormap = {"U": "#e41a1c", "Y": "#377eb8", "Z": "#4daf4a"}


def thresholds(df):
    df = pa.data.transform_errors(df)
    return px.scatter(
        df,
        x="Day",
        y="50%",
        error_y="err+",
        error_y_minus="err-",
        color="Subject",
        color_discrete_map=colormap,
        template=template,
    )


def curves(curve_data):
    df = curve_data["curves_df"]
    y = curve_data["y"]
    if "5%" in df.columns:
        df = pa.data.transform_errors(df)
    df = df.reset_index()
    return px.scatter(
        df,
        x="x",
        y=y,
        error_y="err+",
        error_y_minus="err-",
        color=df.get("Subject"),  # or df["Type"],
        color_discrete_map=colormap,
        symbol=df.get("Day"),
        template=template,
    )


def logistic(data):
    df = data.reset_index()
    return px.line(
        df, x="x", y="Hit Rate", color=df.get("Type"), template="plotly_white"
    )


def bayes(simulated, estimated):
    estimated = estimated.to_frame().rename(columns={"50%": "Hit Rate"})
    df = pd.concat(
        [simulated.reset_index(), estimated.reset_index()],
        keys=["Simulated", "Estimated"],
        names=["Type"],
    )
    return px.scatter(
        df.reset_index(), x="x", y="Hit Rate", color="Type", template="plotly_white"
    )


def hits_animation(cumulative_draws: pd.DataFrame):
    df = cumulative_draws
    return px.bar(
        df.reset_index(),
        x="x",
        y="Hits",
        animation_group="x",
        animation_frame="n",
        range_y=(0, max(df["Hits"])),
    )


def hit_rate_animation(cumulative_draws: pd.DataFrame):
    df = cumulative_draws.reset_index()
    return px.scatter(
        df,
        x="x",
        y="Hit Rate",
        error_y="err+",
        error_y_minus="err-",
        color=df.get("Subject") or df.get("Type"),
        symbol=df.get("Day"),
        animation_group="x",
        animation_frame="n",
        template=template,
    )


def posterior_animation(cumulative_draws: pd.DataFrame):
    df = cumulative_draws
    df = pa.data.transform_errors(df).reset_index()
    return px.scatter(
        df,
        x="x",
        y="Hit Rate",
        error_y="err+",
        error_y_minus="err-",
        animation_group="x",
        animation_frame="n",
        color=df.get("Subject"),
        symbol=df.get("Day"),
        template=template,
    )


def difference_thresholds():
    return px.scatter(
        pd.DataFrame(
            {
                "Reference Charge (nC)": [10.0, 20.0, 30.0],
                "Difference Threshold Charge (nC)": [0.1, 0.2, 0.3],
                "Monkey": ["U", "U", "U"],
                "Dimension": ["PW", "PW", "PW"],
                "n": [10, 7, 9],
            }
        ),
        x="Reference Charge (nC)",
        y="Difference Threshold Charge (nC)",
        color="Monkey",
        color_discrete_map=colormap,
        symbol="Dimension",
        size="n",
        trendline="ols",
        template=template,
    )


labels = {
    "Amp": {
        "x": "Fixed Pulse Width (μs)",
        "y": {
            "inverse": "Threshold Amplitude (μA)",
            "linear": "Threshold Charge (nC)",
        },
    },
    "Width": {
        "x": "Fixed Amplitude (μA)",
        "y": {
            "inverse": "Threshold Pulse Width (μs)",
            "linear": "Threshold Charge (nC)",
        },
    },
}


def get_labels_given_dim(labels, dim, plot_type):
    return {"x": labels[dim]["x"], "y": labels[dim]["y"][plot_type]}


def strength_duration(
    data=None, dim=None, plot_type=None, x_data=[], y_data=[], points=None
):

    labels_given_dim = get_labels_given_dim(labels=labels, dim=dim, plot_type=plot_type)
    x = labels_given_dim["x"]
    y = labels_given_dim["y"]
    if data is not None:
        sd_df = pa.data.filter(data, dim=dim)
    else:
        sd_df = pd.DataFrame({x: x_data, y: y_data})
    return px.scatter(
        sd_df,
        x=x,
        y=y,
        template=template,
    )


def counts(sessions, dim=None):
    if dim is not None:
        sessions["Dimension"] = "Amp"
        sessions = sessions[sessions["Dimension"] == dim]
    return px.histogram(
        sessions,
        x="Monkey",
        color="Monkey",
        template=template,
    ).update_layout(yaxis_title_text="# of Sessions")


def ecdf(blocks, param):
    return px.ecdf(
        blocks.reset_index(), x=param, color=blocks.get("Monkey")
    ).update_layout(xaxis_title=param)


def psychometric(fit, x_range=(-3, 3)):
    s = pa.psi(fit["Threshold"], fit["width"], fit["lambda"], fit["gamma"], x_range)
    return px.line(s, template=pa.plot.template)



import dash_bootstrap_components as dbc  # type: ignore
from dash import html, dcc  # type: ignore
import psychoanalyze as pa
import dash_daq as daq  # type: ignore
import plotly.express as px  # type: ignore
import pandas as pd

defaults = {
    "session": {"sessions": 1, "trials": 100, "subjects": 1},
    "param": {"thresh": 0, "scale": 1, "gamma": 0, "lambda": 0},
    "x_min": {"x_min": -3},
    "x_max": {"x_max": 3},
}


def label(type, param):
    labels = {
        "param": f"{param}",
        "session": f"{param}",
        "x_min": "Min: ",
        "x_max": "Max: ",
    }
    return labels[type]


def input_pair(type, param, default_value):
    _label = label(type, param)
    return dbc.InputGroup(
        [
            dbc.Label(_label),
            dbc.Input(id=param, value=default_value, type="number"),
        ]
    )


def input_group(type):
    return [input_pair(type, key, value) for key, value in defaults[type].items()]


session_inputs = input_group("session")
param_inputs = input_group("param")
x_min_input = input_group("x_min")
x_max_input = input_group("x_max")

threshold_column = dbc.Col(
    [
        dcc.Graph(id="time-thresholds"),
        dbc.Table(id="thresh-data"),
    ]
)


def points_column(data):
    return dbc.Col(
        [
            dcc.Graph(figure=pa.points.plot(data), id="psych-plot"),
        ],
    )


diff_thresh_column = dbc.Col(
    [
        dcc.Graph(id="difference-thresholds"),
        html.P(id="weber-fraction"),
    ],
)
controls = dbc.Col(
    [
        dbc.Card(
            dbc.CardBody(
                [html.H4("Entity Counts", className="card-title")] + session_inputs
            )
        ),
        dbc.Card(
            dbc.CardBody(
                [html.H4("Simulated Params", className="card-title")] + param_inputs
            ),
        ),
        dbc.Card(
            dbc.CardBody(
                [html.H4("Visible Range", className="card-title")]
                + x_min_input
                + x_max_input
            ),
        ),
    ],
    width=2,
)


def simulation_tab(points):
    return dbc.Tab(
        [
            dbc.Row(
                [
                    dbc.Col(
                        [
                            dbc.Label("n trials"),
                            dbc.Input(type="number", value=1, id="n-trials"),
                        ],
                        width=1,
                        align="center",
                    ),
                    dbc.Col(
                        [
                            points_column(points),
                        ],
                        width=6,
                        align="center",
                    ),
                    # dbc.Col(
                    #     [
                    #         dash_table.DataTable(
                    #             points.reset_index().to_dict("records"),
                    #             id="psych-table",
                    #         ),
                    #     ],
                    #     width=1,
                    #     align="center",
                    # ),
                ],
                justify="center",
            ),
            # dbc.Row(pa.plot.strength_duration()),
        ],
        label="Simulation",
    )


sd_plots = [
    dbc.Row(
        [
            dbc.Col(
                dcc.Graph(
                    # figure=pa.plot.strength_duration(
                    #     data=pa.strength_duration.from_blocks(
                    #         pa.blocks.load(), dim=dim
                    #     ),
                    #     dim=dim,
                    #     plot_type="inverse",
                    # )
                    figure=px.scatter()
                )
            )
            for dim in ["Amp", "Width"]
        ]
    )
    for view in ["inverse", "linear"]
]

sessions = pd.read_csv("data/trials.csv")[["Monkey", "Date"]].drop_duplicates()


def detection_tab(experiment_points, blocks):
    return dbc.Tab(
        dbc.Tabs(
            [
                dbc.Tab(
                    dbc.Row(
                        [
                            dbc.Col(
                                dcc.Graph(figure=pa.plot.counts(sessions), id="counts")
                            ),
                            dbc.Col(
                                dcc.Graph(figure=pa.plot.counts(sessions, dim="Width"))
                            ),
                        ],
                    ),
                    label="Counts",
                ),
                dbc.Tab(sd_plots, label="Strength-Duration Plots", tab_id="sd"),
                dbc.Tab(
                    dbc.Row(
                        [
                            dbc.Col(
                                dcc.Graph(
                                    # figure=pa.plot.ecdf(blocks),
                                    id="ecdf_g_l",
                                )
                            ),
                            # dbc.Col(
                            #     dcc.Graph(
                            #         figure=px.ecdf(blocks, x="width"),
                            #         id="ecdf_amp",
                            #     )
                            # ),
                            dbc.Col(
                                dcc.Graph(
                                    # figure=px.ecdf(blocks, x="Threshold"),
                                    id="ecdf_pw"
                                )
                            ),
                        ]
                    ),
                    label="eCDF",
                ),
                dbc.Tab(
                    dbc.Row(
                        [
                            dbc.Col(
                                dcc.Graph(
                                    figure=pa.points.plot(experiment_points),
                                    id="points",
                                )
                            ),
                            dbc.Col(
                                [
                                    dbc.Button("Fit curves", id="fit"),
                                    pa.points.datatable(experiment_points),
                                ],
                                width=2,
                                align="center",
                            ),
                        ]
                    ),
                    label="Single Block",
                ),
            ],
            active_tab="sd",
        ),
        label="Detection",
    )


discrimination_tab = dbc.Tab(
    [
        dbc.Row(
            [
                dbc.Col(
                    daq.BooleanSwitch(
                        on=True,
                        label="Trendline",
                        labelPosition="top",
                        id="trendline",
                    ),
                ),
                dbc.Col(
                    dbc.RadioItems(
                        options=[
                            {"label": "Log Scale", "value": "Log Scale"},
                            {"label": "Linear Scale", "value": "Linear Scale"},
                        ],
                        value="Log Scale",
                        id="log-scale",
                    ),
                ),
                dbc.Col(
                    dbc.RadioItems(
                        options=[
                            {
                                "label": "Group by x values",
                                "value": "Group by x values",
                            },
                            {"label": "Show all blocks", "value": None},
                        ],
                        value=None,
                        id="group-x",
                    )
                ),
                dbc.Col(
                    dbc.RadioItems(
                        options=[
                            {"label": "Histogram", "value": "Histogram"},
                            {"label": "Off", "value": "Off"},
                        ],
                        value="Histogram",
                        id="marginal",
                    )
                ),
                dbc.Col(
                    dbc.RadioItems(
                        options=[
                            {"label": "error bars", "value": "error bars"},
                            {"label": "off", "value": "off"},
                        ],
                        value="error bars",
                        id="error",
                    )
                ),
            ],
        ),
        dbc.Row(
            [
                dbc.Col(dcc.Graph(id="weber"), width=6, id="weber-width"),
            ]
        ),
    ],
    label="Discrimination",
)


def experiment_tab(experiment_points, blocks):
    return dbc.Tab(
        dbc.Tabs(
            [
                detection_tab(experiment_points, blocks),
                discrimination_tab,
            ]
        ),
        label="Experiment",
    )



import psychoanalyze as pa


def load(points):
    points["Reference Charge (nC)"] = points.index.get_level_values(
        "Amp2"
    ) * points.index.get_level_values("Width2")
    detection_points = points[points["Reference Charge (nC)"] == 0]
    detection_points["Dimension"] = pa.blocks.dimensions(detection_points)
    return detection_points



import pandas as pd
import numpy as np
from scipy.stats import logistic  # type: ignore
from scipy.special import logit, expit  # type: ignore
import psychoanalyze as pa
import plotly.express as px  # type: ignore
import os
import pathlib


dims = ["Amp2", "Width2", "Freq2", "Dur2", "Active Channels", "Return Channels"]
index_levels = dims


def add_posterior(data, posterior):
    return pd.concat(
        [data, posterior],
        keys=["Observed", "Posterior"],
        names=["Type"],
    ).reset_index()


def generate(n_trials_per_level=100):
    index = pd.Index(range(-3, 4), name="x")
    n = [n_trials_per_level] * len(index)
    p = logistic.cdf(index)
    return pd.DataFrame({"n": n, "Hits": np.random.binomial(n, p)}, index=index)


def hit_rate(df: pd.DataFrame) -> pd.Series:
    return df["Hits"] / df["n"]


def xrange_index(x_min, x_max):
    return pd.Index(list(range(x_min, x_max + 1)), name="x")


def transform(hit_rate, y: str):
    return logit(hit_rate) if y == "alpha" else hit_rate


def prep_psych_curve(curves_data: pd.DataFrame, x: pd.Index, y: str):
    curves_data.index = x
    df = pa.points.fit(curves_data)
    df = pa.data.reshape_fit_results(df, x, y)
    return df


def get_fit_param(fit: pd.DataFrame, name: str):
    return fit.loc[name, "50%"]


def from_points(points: pd.DataFrame, dim=None):
    levels = pa.schemas.block_index_levels
    if dim == "Amp":
        levels = levels + ["Width1"]
        df = points.groupby(levels).apply(pa.points.to_block)
        return df[df["n Levels"] > 1].drop(columns="Dimension")
    return points.groupby(levels).apply(pa.points.to_block)


def from_trials(trials: pd.DataFrame) -> pd.Series:
    points = pa.points.from_trials(trials)
    return from_points(points)


def dimensions(points, dims=None):
    if dims is None:
        return pa.points.dimension(points)
    return points.groupby(
        [dim for dim in list(points.index.names) if dim not in dims]
    ).apply(pa.points.dimension)


def fits(points):
    if len(points):
        return points.groupby(pa.schemas.block_index_levels).apply(pa.points.fit)
    else:
        return pd.DataFrame(
            {"Threshold": [], "Fixed Magnitude": [], "Dimension": []},
            index=pd.MultiIndex.from_frame(
                pd.DataFrame(
                    {
                        "Monkey": [],
                        "Date": [],
                        "Amp2": [],
                        "Width2": [],
                        "Freq2": [],
                        "Dur2": [],
                        "Active Channels": [],
                        "Return Channels": [],
                    }
                )
            ),
        )


def empty():
    return pd.Series(
        [], name="Hit Rate", index=pd.Index([], name="Amplitude (µA)"), dtype=float
    )


def plot_fits(df):
    x = np.linspace(-3, 3, 100)
    y = expit(x)
    return px.line(df.reset_index(), x=x, y=y)


def load_cached(data_path):
    channel_config = ["Active Channels", "Return Channels"]
    blocks = pd.read_csv(data_path / "blocks.csv", parse_dates=["Date"]).set_index(
        pa.sessions.dims + pa.stimulus.ref_dims + channel_config
    )
    blocks["Day"] = days(blocks, pa.subjects.load(data_path))
    return blocks


def load(data_path=pathlib.Path("data"), monkey=None, day=None, dim=None):
    blocks_path = data_path / "blocks.csv"
    if os.path.exists(blocks_path):
        blocks = load_cached(data_path)
        if monkey:
            if day:
                blocks = blocks[blocks["Day"] == day]
            else:
                blocks = blocks.xs(monkey, drop_level=False)
        blocks = blocks[blocks["n Levels"] > 1]
        return blocks
    else:
        blocks = from_trials(pa.trials.load(data_path))
        blocks.to_csv(blocks_path)
        return blocks


def fixed_magnitudes(points):
    return points.groupby(pa.schemas.block_index_levels).agg(pa.points.fixed_magnitude)


def days(blocks, intervention_dates):
    blocks = blocks.join(intervention_dates, on="Monkey")
    days = (blocks.index.get_level_values("Date") - blocks["Surgery Date"]).dt.days
    days.name = "Days"
    return days


def n_trials(trials):
    session_cols = ["Monkey", "Date"]
    ref_stim_cols = ["Amp2", "Width2", "Freq2", "Dur2"]
    channel_config = ["Active Channels", "Return Channels"]
    return trials.groupby(session_cols + ref_stim_cols + channel_config)[
        "Result"
    ].count()


def read_fit(path, block):
    session_cols = ["Monkey", "Date"]
    ref_stim_cols = ["Amp2", "Width2", "Freq2", "Dur2"]
    channel_config = ["Active Channels", "Return Channels"]
    fits = pd.read_csv(
        path,
        index_col=session_cols + ref_stim_cols + channel_config,
        parse_dates=["Date"],
    )
    fits["err+"] = 0.0
    fits["err-"] = 0.0
    if block in fits.index:
        return fits.loc[block]
    else:
        return pd.Series()


def experiment_type(blocks):
    ref_stim = blocks.reset_index()[["Amp2", "Width2", "Freq2", "Dur2"]]
    ref_charge = ref_stim["Amp2"] * ref_stim["Width2"]
    blocks.loc[ref_charge == 0, "Experiment Type"] = "Detection"
    blocks.loc[ref_charge != 0, "Experiment Type"] = "Discrimination"
    return blocks["Experiment Type"]


def isValid(block):
    return any(block["Hit Rate"] > 0.5) & any(block["Hit Rate"] < 0.5)


def monkey_counts(data):
    summary = (
        data.index.get_level_values("Monkey").value_counts().rename("Total Blocks")
    )
    summary.index.name = "Monkey"
    return summary



import pandas as pd
import numpy as np
import psychoanalyze as pa
from pathlib import Path
import random
from datetime import datetime
import json


data_path = Path("data/trials.csv")


def generate_hits(n, p):
    return np.random.binomial(n, p)


def generate_block(dim="x"):
    hits = pd.Series(
        [generate_hits(100, p) for p in [0.1, 0.2, 0.5, 0.8, 0.9]], name="Hits"
    )
    x = [0, 1, 2, 3, 4]
    return pd.DataFrame({"Hits": hits, "n": [100] * 5}, index=pd.Index(x, name=dim))


def generate(n, stim_levels=None):
    return pa.schemas.trials.validate(
        pd.DataFrame(
            {"Result": np.random.binomial(1, 0.5, n)},
            index=pd.MultiIndex.from_frame(
                pd.DataFrame(
                    {
                        "Monkey": ["Z"] * n,
                        "Date": [datetime.now()] * n,
                        "Amp2": [0.0] * n,
                        "Width2": [0.0] * n,
                        "Freq2": [0.0] * n,
                        "Dur2": [0.0] * n,
                        "Active Channels": [1] * n,
                        "Return Channels": [1] * n,
                        "Amp1": random.choices(list(range(8)), k=n),
                        "Width1": [0.0] * n,
                        "Freq1": [0.0] * n,
                        "Dur1": [0.0] * n,
                    }
                )
            ),
        )
    )


def load(data_path: Path = Path("data")):
    return pa.schemas.trials.validate(
        pd.read_csv(
            data_path / "trials.csv",
            index_col=pa.schemas.points_index_levels,
            parse_dates=["Date"],
        )
    )


def from_store(store_data):
    df_dict = json.loads(store_data)
    index_names = df_dict.pop("index_names")
    index = pd.MultiIndex.from_tuples(df_dict["index"])
    df = pd.DataFrame({"Result": df_dict["data"][0]}, index=index)
    df.index.names = index_names
    return pa.schemas.trials.validate(df)


def to_store(df):
    df.index = df.index.set_levels(df.index.levels[1].astype(str), level=1)
    data_dict = df.to_dict(orient="split")
    data_dict["index_names"] = pa.schemas.points_index_levels
    return json.dumps(data_dict)


def normalize(trials):
    return {
        "Session": trials[["Monkey", "Day"]].drop_duplicates(),
        "Reference Stimulus": trials[["Amp2", "Width2", "Freq2", "Dur2"]],
        "Channel Config": trials[["Active Channels", "Return Channels"]],
        "Test Stimulus": trials[["Amp1", "Width1", "Freq1", "Dur1"]],
    }



import pandera as pr

session_dims = ["Monkey", "Date"]
block_stim_dims = ["Amp2", "Width2", "Freq2", "Dur2"]
block_channel_dims = ["Active Channels", "Return Channels"]
block_dims = block_stim_dims + block_channel_dims
point_dims = ["Amp1", "Width1", "Freq1", "Dur1"]

block_index_levels = session_dims + block_dims
points_index_levels = block_index_levels + point_dims


points = pr.DataFrameSchema(
    {
        "n": pr.Column(int, coerce=True),
        "x": pr.Column(float, coerce=True),
        "Hits": pr.Column(float, coerce=True),
    },
    index=pr.MultiIndex(
        [
            pr.Index(str, name="Monkey"),
            pr.Index("datetime64", name="Date", coerce=True),
        ]
        + [pr.Index(float, name=dim) for dim in block_stim_dims]
        + [pr.Index(int, name=dim) for dim in block_channel_dims]
        + [pr.Index(float, name=dim) for dim in point_dims]
    ),
    coerce=True,
)

blocks = pr.DataFrameSchema(
    columns={"Threshold": pr.Column(dtype=float), "width": pr.Column(dtype=float)},
    index=pr.MultiIndex(
        [
            pr.Index(str, name="Monkey"),
            pr.Index("datetime64", name="Date", coerce=True),
        ]
        + [pr.Index(float, name=dim) for dim in block_stim_dims]
        + [pr.Index(int, name=dim) for dim in block_channel_dims]
    ),
)


trials = pr.DataFrameSchema(
    {"Result": pr.Column(int, coerce=True)},
    index=pr.MultiIndex(
        [
            pr.Index(str, name="Monkey"),
            pr.Index("datetime64", name="Date", coerce=True),
        ]
        + [pr.Index(float, name=dim) for dim in block_stim_dims]
        + [pr.Index(int, name=dim) for dim in block_channel_dims]
        + [pr.Index(float, name=dim) for dim in point_dims]
    ),
    coerce=True,
)



import pandas as pd
import cmdstanpy as stan
import plotly.express as px  # type: ignore
from scipy.stats import binom  # type: ignore
import psychoanalyze as pa
from dash import dash_table  # type: ignore
import pathlib
from plotly import graph_objects as go
import numpy as np


index_levels = ["Amp1", "Width1", "Freq1", "Dur1"]


def from_trials(trials):
    trials = trials[trials["Result"].isin([0, 1])]
    points = (
        trials.groupby(trials.index.names)["Result"]
        .agg(["count", "sum"])
        .rename(columns={"count": "n", "sum": "Hits"})
    )
    points["x"] = points.index.get_level_values("Amp1")
    return points


def load(data_path=pathlib.Path("data")):
    trials = pa.trials.load(data_path)
    points = from_trials(trials)
    points["Hit Rate"] = points["Hits"] / points["n"]
    return points


def dimension(points):
    amp1, width1 = (
        points.index.get_level_values(param) for param in ["Amp1", "Width1"]
    )
    if amp1.nunique() > 1 and width1.nunique() == 1:
        return "Amp"
    elif width1.nunique() > 1 and amp1.nunique() == 1:
        return "Width"
    elif width1.nunique() > 1 and amp1.nunique() > 1:
        return "Both"


def prep_fit(points: pd.DataFrame, dimension="Amp1"):
    points = points.reset_index()
    return {
        "X": len(points),
        "x": points[f"{dimension}"].to_numpy(),
        "N": points["n"].to_numpy(),
        "hits": points["Hits"].to_numpy(),
    }


def model():
    return stan.CmdStanModel(stan_file="models/binomial_regression.stan")


# def fit(ready_for_fit: pd.DataFrame) -> pd.DataFrame:
#     _model = model()
#     return _model.sample(chains=4, data=ready_for_fit)


def fit(points, save_to=None, block=None):
    points = points[["x", "Hits", "n"]]
    if len(points):

        data = points.to_numpy()
        fit = pa.fit(data)
        fit = pd.DataFrame(
            {
                "Threshold": [fit["Fit"][0]],
                "width": [fit["Fit"][1]],
                "gamma": [fit["Fit"][2]],
                "lambda": [fit["Fit"][3]],
                "err+": [None],
                "err-": [None],
            },
            index=pd.MultiIndex.from_tuples(
                [block],
                names=[
                    "Monkey",
                    "Date",
                    "Amp2",
                    "Width2",
                    "Freq2",
                    "Dur2",
                    "Active Channels",
                    "Return Channels",
                ],
            ),
        )
        if save_to:
            fit.to_csv(save_to)
        return fit
    else:
        return pd.Series(
            {
                "Threshold": None,
                "width": None,
                "lambda": None,
                "gamma": None,
                "err+": None,
                "err-": None,
            }
        )


def plot(points, trendline=None):
    points["Hit Rate"] = points["Hits"] / points["n"]
    return px.scatter(
        points.reset_index(),
        x="x",
        y="Hit Rate",
        size="n",
        color=points.get("Monkey"),
        template=pa.plot.template,
        trendline=trendline,
    )


def generate(x, n, p):
    return pd.Series(
        [binom.rvs(n[i], p[i]) for i in range(len(x))],
        index=pd.Index(x, name="Amplitude (µA)"),
        name="Hit Rate",
    )


def datatable(data):
    table = dash_table.DataTable(
        data.reset_index()[["Amp1", "Hit Rate", "n"]].to_dict("records"),
        columns=[
            {
                "id": "Amp1",
                "name": "Amp1",
                "type": "numeric",
                "format": dash_table.Format.Format(
                    precision=2, scheme=dash_table.Format.Scheme.fixed
                ),
            },
            {
                "id": "Hit Rate",
                "name": "Hit Rate",
                "type": "numeric",
                "format": dash_table.Format.Format(
                    precision=2, scheme=dash_table.Format.Scheme.fixed
                ),
            },
            {
                "id": "n",
                "name": "n",
                "type": "numeric",
            },
        ],
        id="experiment-psych-table",
    )
    return table


def from_store(store_data):
    trials = pa.trials.from_store(store_data)
    return from_trials(trials)


def combine_plots(fig1, fig2):
    return go.Figure(data=fig1.data + fig2.data)


def fixed_magnitude(points):
    dim = dimension(points)
    if dim == "Amp":
        return points.index.get_level_values("Width1")[0]
    if dim == "Width":
        return points.index.get_level_values("Amp1")[0]


def n(points):
    return len(points)


def to_block(points):
    return pd.Series(
        {
            # "Threshold": fit(points),
            "Dimension": dimension(points),
            "Fixed Magnitude": fixed_magnitude(points),
            "n Levels": n(points),
        }
    )


def psi(x: np.ndarray, threshold: float, width: float, gamma: float, lambda_: float):
    return gamma + (1 - gamma - lambda_) / (
        1 + np.exp(-gamma * (x - threshold) / width) ** lambda_
    )



import plotly.express as px  # type: ignore
import psychoanalyze as pa
import pandas as pd


def plot(
    data,
    trendline="ols",
    group_x=None,
    log_scale=True,
    marginal=None,
    error_y=None,
    stable_sessions=None,
):
    if stable_sessions == "stable":
        data = data[data["Date"].between("6/19/2017", "8/16/2017")]
    if group_x:
        data = pa.weber.aggregate(data)
    _log_scale = True if log_scale == "Log Scale" else None
    _marginal = "histogram" if marginal == "Histogram" else None
    _trendline = "ols" if trendline else None
    return px.scatter(
        data.reset_index(),
        x="Reference Charge (nC)",
        y="Difference Threshold (nC)",
        error_y="err+" if error_y == "error bars" else None,
        error_y_minus="err-" if error_y == "error bars" else None,
        color="Monkey",
        color_discrete_map=pa.plot.colormap,
        symbol="Dimension",
        trendline=_trendline,
        template="plotly_white",
        log_x=_log_scale,
        trendline_options={"log_x": _log_scale},
        marginal_x=_marginal,
        hover_data=["Date"],
    )


def aggregate(data):
    return (
        data.groupby(["Monkey", "Dimension", "Reference Charge (nC)"])[
            "Difference Threshold (nC)"
        ]
        .agg(["mean", "count", "std"])
        .rename(columns={"mean": "Difference Threshold (nC)"})
    )


def load(path):
    df = pd.read_csv(path, parse_dates=["Date"])
    df["err+"] = (df["location_CI_5"] * df["Fixed_Param_Value"] / 1000) - df[
        "Threshold_Charge_nC"
    ]
    df["err-"] = (
        df["Threshold_Charge_nC"]
        - (df["location_CI_95"]) * df["Fixed_Param_Value"] / 1000
    )
    return df



import os
from typing import List
import pandas as pd
import pathlib

import psychoanalyze as pa


dims = ["Monkey", "Date"]
index_levels = dims


def generate(n: int) -> List[int]:
    return list(range(n))


def from_trials_csv(path, cache=False):
    df = pd.read_csv(path)[["Monkey", "Date"]].drop_duplicates()
    if cache:
        df.to_csv("data/normalized/sessions.csv", index=False)
    return df


def day_marks(subjects, sessions, monkey):
    surgery_date = pd.to_datetime(
        subjects.loc[subjects["Monkey"] == monkey, "Surgery Date"]
    )[0]
    sessions = sessions[sessions["Monkey"] == "U"]
    sessions["Days"] = (pd.to_datetime(sessions["Date"]) - surgery_date).dt.days
    return {sessions.loc[i, "Days"]: sessions.loc[i, "Date"] for i in sessions.index}


def cache():
    pd.read_csv("data/trials.csv")[["Monkey", "Date"]].drop_duplicates().to_csv(
        "data/normalized/sessions.csv", index=False
    )


def load(data_path=pathlib.Path("data"), monkey=None):
    if os.path.exists(data_path / "sessions.csv"):
        sessions = pd.read_csv(data_path / "sessions.csv")
    else:
        trials = pa.trials.load(data_path)
        if monkey:
            trials = trials[trials.index.get_level_values("Monkey") == monkey]
        sessions = (
            trials.groupby(["Monkey", "Date"])[["Result"]]
            .count()
            .rename(columns={"Result": "n Trials"})
        )
    sessions["Day"] = days(sessions, pa.subjects.load(data_path))
    return sessions


def days(sessions: pd.DataFrame, subjects):
    df = sessions.join(subjects, on="Monkey")
    return (
        pd.to_datetime(df.index.get_level_values("Date")) - df["Surgery Date"]
    ).dt.days


def n_trials(sessions, trials):
    return trials.groupby(["Monkey", "Date"])[["Result"]].count()


def load_cached(data_dir, monkey=None):
    sessions = pd.read_csv(data_dir / "sessions.csv", index_col=["Monkey", "Date"])
    return sessions[sessions.index.get_level_values("Monkey") == monkey]



import psychoanalyze as pa


def xlabel():
    pass


def from_blocks(blocks, dim=None):
    if dim == "Amp":
        ylabel = "Threshold Amplitude (μA)"
        xlabel = "Fixed Pulse Width (μs)"
    elif dim == "Width":
        ylabel = "Fixed Amplitude (μA)"
        xlabel = "Threshold Pulse Width (μs)"

    blocks[ylabel] = blocks["Threshold"]
    blocks[xlabel] = blocks["Fixed Magnitude"]
    return blocks.drop(columns=["Threshold", "Fixed Magnitude"])


def plot(plot_type, dim=None):
    return pa.plot.strength_duration(dim=dim, plot_type=plot_type)



import pandas as pd
import numpy as np
from psychoanalyze import (
    schemas,
    sessions,
    blocks,
    points,
    plot,
    data,
    trials,
    weber,
    detection,
    amp,
    strength_duration,
    subjects,
    stimulus,
    simulate,
)

pd.options.plotting.backend = "plotly"

__version__ = "0.1.0"

__all__ = [
    "schemas",
    "plot",
    "data",
    "trial",
    "blocks",
    "sessions",
    "weber",
    "detection",
    "amp",
    "points",
    "trials",
    "blocks",
    "strength_duration",
    "subjects",
    "stimulus",
    "simulate",
]

# def curve(trials: pd.DataFrame) -> pd.Series:
#     """Arrange *method of constant stimuli* performance curves using trial data"""
#     df = trials.groupby("x").mean().rename(columns={"Result": "Hit Rate"})
#     return df


def weber_coefficient(curves: pd.DataFrame) -> float:
    """Calculate weber coefficient for a set of psychometric curves"""
    return 1


def psi(threshold=0, slope=1, lambda_=0, gamma=0, x_range=(-3, 3)) -> pd.Series:
    """Basic sigmoid psychometric function psi (Ψ) = expit/logistic"""
    x = np.linspace(x_range[0], x_range[1])
    y = gamma + (1 - lambda_ - gamma) * 1 / (1 + np.exp((-slope) * (x - threshold)))
    # y = gamma + (1 - lambda_ - gamma) * expit(x)
    return pd.Series(y, index=x)  # type: ignore


def fit(data):
    return {"Fit": [1, 1, 0, 0]}
